# -*- coding: utf-8 -*-
##############################################################################
#    
#    VNC Developments (India) Pvt. Ltd.
#    Copyright (C) 2004-TODAY VNC (<http://www.vnc.biz>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.     
#
##############################################################################

from osv import osv
from osv import fields
import netsvc
from tools.translate import _
import time

class mrp_product_produce(osv.osv):
    _inherit = 'mrp.product.produce'

    _columns = {
        'prodlot_id':fields.many2one('stock.production.lot','Production Lot'),
        'product_id': fields.many2one('product.product', 'Product', ),
    }
    def _get_prodlot(self, cr, uid, context):
        """docstring for _get_prodlot"""
        ret_val = False
        if not context.get('active_id',False):
            return ret_val
        mrp_obj = self.pool.get('mrp.production').browse(cr, uid,
                                               context['active_id'],
                                               context=context)
        for line in mrp_obj.move_created_ids:
            if line.prodlot_id:
                ret_val = line.prodlot_id.id
        return ret_val

    def _get_product_id(self, cr, uid, context):
        """docstring for _get_prodlot"""
        ret_val = False
        if not context.get('active_id',False):
            return ret_val
        mrp_obj = self.pool.get('mrp.production').browse(cr, uid,
                                               context['active_id'],
                                               context=context)
        return mrp_obj.product_id.id

    _defaults = {
         'prodlot_id': _get_prodlot,
         'product_id': _get_product_id,
    }
    def do_produce(self, cr, uid, ids, context=None):
        data = self.browse(cr, uid, ids[0], context=context)
        mrp_obj = self.pool.get('mrp.production').browse(cr, uid,
                                               context['active_id'],
                                               context=context)

        self.pool.get('stock.move').write(cr, uid ,[x.id for x in mrp_obj.move_created_ids],
                                          {'prodlot_id':data.prodlot_id.id})
        return super(mrp_product_produce, self).do_produce(cr, uid, ids, context=context)

mrp_product_produce()
class mrp_production(osv.osv):
    _inherit = 'mrp.production'


    def _make_production_consume_line(self, cr, uid, production_line, parent_move_id, source_location_id=False, context=None):
        stock_move = self.pool.get('stock.move')
        production = production_line.production_id
        # Internal shipment is created for Stockable and Consumer Products
        if production_line.product_id.type not in ('product', 'consu'):
            return False
        move_name = _('PROD: %s') % production.name
        destination_location_id = production.product_id.product_tmpl_id.property_stock_production.id
        if not source_location_id:
            source_location_id = production.location_src_id.id

        move_ids = []
        move_id = stock_move.create(cr, uid, {
            'name': move_name,
            'date': production.date_planned,
            'product_id': production_line.product_id.id,
            'product_qty': production_line.product_qty,
            'prodlot_id': production_line.prodlot_id.id,
            'product_uom': production_line.product_uom.id,
            'product_uos_qty': production_line.product_uos and production_line.product_uos_qty or False,
            'product_uos': production_line.product_uos and production_line.product_uos.id or False,
            'location_id': source_location_id,
            'location_dest_id': destination_location_id,
            'move_dest_id': parent_move_id,
            'state': 'waiting',
            'company_id': production.company_id.id,
        })
        production.write({'move_lines': [(4, move_id)]}, context=context)
        return move_id

    def action_compute(self, cr, uid, ids, properties=[], context=None):
#        res = super(mrp_production, self).action_compute(cr, uid ,ids, properties, context)
        """ Computes bills of material of a product.
        @param properties: List containing dictionaries of properties.
        @return: No. of products.
        """
        results = []
        bom_obj = self.pool.get('mrp.bom')
        uom_obj = self.pool.get('product.uom')
        prod_line_obj = self.pool.get('mrp.production.product.line')
        workcenter_line_obj = self.pool.get('mrp.production.workcenter.line')
        for production in self.browse(cr, uid, ids):
            cr.execute('delete from mrp_production_product_line where production_id=%s', (production.id,))
            cr.execute('delete from mrp_production_workcenter_line where production_id=%s', (production.id,))
            bom_point = production.bom_id
            bom_id = production.bom_id.id
            if not bom_point:
                bom_id = bom_obj._bom_find(cr, uid, production.product_id.id, production.product_uom.id, properties)
                if bom_id:
                    bom_point = bom_obj.browse(cr, uid, bom_id)
                    routing_id = bom_point.routing_id.id or False
                    self.write(cr, uid, [production.id], {'bom_id': bom_id, 'routing_id': routing_id})

            if not bom_id:
                raise osv.except_osv(_('Error'), _("Couldn't find a bill of material for this product."))
            factor = uom_obj._compute_qty(cr, uid, production.product_uom.id, production.product_qty, bom_point.product_uom.id)
            res = bom_obj._bom_explode(cr, uid, bom_point, factor / bom_point.product_qty, properties, routing_id=production.routing_id.id)
            results = res[0]
            results2 = res[1]
            for line in results:
                line['production_id'] = production.id

                res = self.pool.get('product.product').get_stock_controling(cr, uid, line['product_id'],
                                                                      line['product_uom'],
                                                                      line['product_qty'],  production.location_src_id.id)
                if not res:
                    res = [[line['product_qty'],False]]
                total = sum([r[-1] for r in res])
                if total != line['product_qty']:
                    res.append([False,line['product_qty']-total])
                for prodlot_id,qty in res:
                    new_line = line.copy()
                    new_line['product_qty'] = qty
                    new_line['prodlot_id'] = prodlot_id
                    prod_line_obj.create(cr, uid, new_line)
            for line in results2:
                line['production_id'] = production.id
                workcenter_line_obj.create(cr, uid, line)
        return len(results)


    def action_produce(self, cr, uid, production_id, production_qty, production_mode, context=None):
        """ To produce final product based on production mode (consume/consume&produce).
        If Production mode is consume, all stock move lines of raw materials will be done/consumed.
        If Production mode is consume & produce, all stock move lines of raw materials will be done/consumed
        and stock move lines of final product will be also done/produced.
        @param production_id: the ID of mrp.production object
        @param production_qty: specify qty to produce
        @param production_mode: specify production mode (consume/consume&produce).
        @return: True
        """
        stock_mov_obj = self.pool.get('stock.move')
        production = self.browse(cr, uid, production_id, context=context)

        produced_qty = 0
        for produced_product in production.move_created_ids2:
            if (produced_product.scrapped) or (produced_product.product_id.id <> production.product_id.id):
                continue
            produced_qty += produced_product.product_qty

        if production_mode in ['consume','consume_produce']:
            consumed_data = {}

            # Calculate already consumed qtys
            for consumed in production.move_lines2:
                if consumed.scrapped:
                    continue
                if not consumed_data.get(consumed.product_id.id, False):
                    consumed_data[consumed.product_id.id] = 0
                consumed_data[consumed.product_id.id] += consumed.product_qty

            # Find product qty to be consumed and consume it
            com_raw_pr = []
            for scheduled in production.product_lines:

                # total qty of consumed product we need after this consumption
                total_consume = ((production_qty + produced_qty) * scheduled.product_qty / production.product_qty)

                # qty available for consume and produce
                qty_avail = scheduled.product_qty - consumed_data.get(scheduled.product_id.id, 0.0)
                if qty_avail <= 0.0:
                    # there will be nothing to consume for this raw material
                    continue
                raw_product = [move for move in production.move_lines if move.product_id.id==scheduled.product_id.id and move.id not in com_raw_pr]
                if raw_product:
                    # qtys we have to consume
                    qty = total_consume - consumed_data.get(scheduled.product_id.id, 0.0)

                    if qty > qty_avail:
                        # if qtys we have to consume is more than qtys available to consume
                        prod_name = scheduled.product_id.name_get()[0][1]
                        raise osv.except_osv(_('Warning!'), _('You are going to consume total %s quantities of "%s".\nBut you can only consume up to total %s quantities.') % (qty, prod_name, qty_avail))
                    if qty <= 0.0:
                        # we already have more qtys consumed than we need
                        continue
                    raw_product[0].action_consume(qty, raw_product[0].location_id.id, context=context)
                    com_raw_pr.append(raw_product[0].id)

        if production_mode == 'consume_produce':
            # To produce remaining qty of final product
            #vals = {'state':'confirmed'}
            #final_product_todo = [x.id for x in production.move_created_ids]
            #stock_mov_obj.write(cr, uid, final_product_todo, vals)
            #stock_mov_obj.action_confirm(cr, uid, final_product_todo, context)
            produced_products = {}
            for produced_product in production.move_created_ids2:
                if produced_product.scrapped:
                    continue
                if not produced_products.get(produced_product.product_id.id, False):
                    produced_products[produced_product.product_id.id] = 0
                produced_products[produced_product.product_id.id] += produced_product.product_qty

            for produce_product in production.move_created_ids:
                produced_qty = produced_products.get(produce_product.product_id.id, 0)
                subproduct_factor = self._get_subproduct_factor(cr, uid, production.id, produce_product.id, context=context)
                rest_qty = (subproduct_factor * production.product_qty) - produced_qty

                if rest_qty < production_qty:
                    prod_name = produce_product.product_id.name_get()[0][1]
                    raise osv.except_osv(_('Warning!'), _('You are going to produce total %s quantities of "%s".\nBut you can only produce up to total %s quantities.') % (production_qty, prod_name, rest_qty))
                if rest_qty > 0 :
                    stock_mov_obj.action_consume(cr, uid, [produce_product.id], (subproduct_factor * production_qty), context=context)

        for raw_product in production.move_lines2:
            new_parent_ids = []
            parent_move_ids = [x.id for x in raw_product.move_history_ids]
            for final_product in production.move_created_ids2:
                if final_product.id not in parent_move_ids:
                    new_parent_ids.append(final_product.id)
            for new_parent_id in new_parent_ids:
                stock_mov_obj.write(cr, uid, [raw_product.id], {'move_history_ids': [(4,new_parent_id)]})

        wf_service = netsvc.LocalService("workflow")
        wf_service.trg_validate(uid, 'mrp.production', production_id, 'button_produce_done', cr)
        return True

    def action_production_end(self, cr, uid, ids):
        """ Changes production state to Finish and writes finished date.
        @return: True
        """
        def get_value(line):
            if not line.prodlot_id:
                return 0.0
            move_obj = [x for x in line.prodlot_id.move_ids if (x.picking_id and x.picking_id.type=='in' or x.production_id ) and x.lot_price > 0]
            if not move_obj:
                return 0.0
            return move_obj[0].lot_price * line.product_qty
        for production in self.browse(cr, uid, ids):
            amount = self._costs_generate(cr, uid, production)
            for line in production.move_lines2:
                amount+= get_value(line)
            per_price = amount / production.product_qty
            self.pool.get('stock.move').write(cr, uid, [line.id for line in production.move_created_ids2],{'lot_price':per_price})
        return self.write(cr, uid, ids, {'state': 'done', 'date_finished': time.strftime('%Y-%m-%d %H:%M:%S')})

mrp_production()


class mrp_production_product_line(osv.osv):
    _inherit = 'mrp.production.product.line'

    _columns = {
        'prodlot_id':fields.many2one('stock.production.lot','Production Lot'),
    }

mrp_production_product_line()

